=== modified file 'bzrlib/tests/per_transport.py'
--- bzrlib/tests/per_transport.py	2010-02-23 07:43:11 +0000
+++ bzrlib/tests/per_transport.py	2010-02-25 03:23:48 +0000
@@ -1083,6 +1083,52 @@
         subdir.stat('./file')
         subdir.stat('.')
 
+    def test_link(self):
+        from stat import ST_NLINK
+
+        t = self.get_transport()
+
+        source_name = "original_target"
+        link_name = "target_link"
+
+        self.build_tree([source_name], transport=t)
+
+        try:
+            link_result = t.link(source_name, link_name)
+            self.failUnless(link_result)
+
+            self.failUnless(t.has(source_name))
+            self.failUnless(t.has(link_name))
+
+            st = t.stat(link_name)
+            self.failUnlessEqual(st[ST_NLINK], 2)
+        except TransportNotPossible:
+            # This transport doesn't do hardlinks
+            return
+
+    def test_symlink(self):
+        from stat import S_ISLNK
+
+        t = self.get_transport()
+
+        source_name = "original_target"
+        link_name = "target_link"
+
+        self.build_tree([source_name], transport=t)
+
+        try:
+            link_result = t.symlink(source_name, link_name)
+            self.failUnless(link_result)
+
+            self.failUnless(t.has(source_name))
+            self.failUnless(t.has(link_name))
+
+            st = t.stat(link_name)
+            self.failUnless(S_ISLNK(st.st_mode))
+        except TransportNotPossible:
+            # This transport doesn't do hardlinks
+            return
+
     def test_list_dir(self):
         # TODO: Test list_dir, just try once, and if it throws, stop testing
         t = self.get_transport()

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2010-02-23 07:43:11 +0000
+++ bzrlib/transport/__init__.py	2010-02-25 03:23:48 +0000
@@ -1202,6 +1202,20 @@
         count = self._iterate_over(relpaths, gather, pb, 'stat', expand=False)
         return stats
 
+    def link(self, source, link_name):
+        """Create a hardlink pointing to source named link_name.
+
+        Return True if successful.
+        """
+        raise errors.TransportNotPossible("Hard links are not supported on %s" % self)
+
+    def symlink(self, source, link_name):
+        """Create a symlink pointing to source named link_name.
+
+        Return True if successful.
+        """
+        raise errors.TransportNotPossible("Symlinks are not supported on %s" % self)
+
     def listable(self):
         """Return True if this store supports listing."""
         raise NotImplementedError(self.listable)

=== modified file 'bzrlib/transport/local.py'
--- bzrlib/transport/local.py	2010-02-23 07:43:11 +0000
+++ bzrlib/transport/local.py	2010-02-25 03:23:48 +0000
@@ -481,7 +481,7 @@
         path = relpath
         try:
             path = self._abspath(relpath)
-            return os.stat(path)
+            return os.lstat(path)
         except (IOError, OSError),e:
             self._translate_error(e, path)
 
@@ -515,6 +515,33 @@
         except (IOError, OSError),e:
             self._translate_error(e, path)
 
+    def link(self, source, link_name):
+        """See Transport.link."""
+        try:
+            os.link(self._abspath(source),
+                    self._abspath(link_name))
+            return True
+        except AttributeError:
+            raise errors.TransportNotPossible("Hardlinks are not supported on %s" % self)
+        except (IOError, OSError), e:
+            self._translate_error(e, source)
+
+    def symlink(self, source, link_name):
+        """See Transport.symlink."""
+        abs_link_dirpath = urlutils.dirname(self.abspath(link_name))
+        source_rel = urlutils.file_relpath(
+            urlutils.strip_trailing_slash(abs_link_dirpath),
+            urlutils.strip_trailing_slash(self.abspath(source))
+        )
+
+        try:
+            os.symlink(source_rel, self._abspath(link_name))
+            return True
+        except AttributeError:
+            raise errors.TransportNotPossible("Symlinks are not supported on %s" % self)
+        except (IOError, OSError), e:
+            self._translate_error(e, source_rel)
+
     def _can_roundtrip_unix_modebits(self):
         if sys.platform == 'win32':
             # anyone else?

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2010-02-23 07:43:11 +0000
+++ bzrlib/transport/sftp.py	2010-02-25 03:23:48 +0000
@@ -82,7 +82,7 @@
 else:
     from paramiko.sftp import (SFTP_FLAG_WRITE, SFTP_FLAG_CREATE,
                                SFTP_FLAG_EXCL, SFTP_FLAG_TRUNC,
-                               CMD_HANDLE, CMD_OPEN)
+                               SFTP_OK, CMD_HANDLE, CMD_OPEN)
     from paramiko.sftp_attr import SFTPAttributes
     from paramiko.sftp_file import SFTPFile
 
@@ -810,10 +810,20 @@
         """Return the stat information for a file."""
         path = self._remote_path(relpath)
         try:
-            return self._get_sftp().stat(path)
+            return self._get_sftp().lstat(path)
         except (IOError, paramiko.SSHException), e:
             self._translate_io_exception(e, path, ': unable to stat')
 
+    def symlink(self, source, link_name):
+        """See Transport.symlink."""
+        try:
+            conn = self._get_sftp()
+            retval = conn.symlink(source, link_name)
+            return (SFTP_OK == retval)
+        except (IOError, paramiko.SSHException), e:
+            self._translate_io_exception(e, link_name,
+                                         ': unable to create symlink to %r' % (source))
+
     def lock_read(self, relpath):
         """
         Lock the given file for shared (read) access.

